/*
 * EukaryoteFilter.cpp
 *
 *  Created on: 2013/01/09
 *      Author: AxioHelix
 */

#include	<string>

#include	<iostream>
#include	<fstream>
#include	<sstream>
#include	<map>
#include	<vector>
#include	<algorithm>
#include	<cmath>

#include	<stdio.h>
#include	<stdlib.h>

#include	"Splitter.h"

#include	"ParameterObj.h"
#include	"BlastResult.h"
#include	"TaxonomyNode.h"
#include	"KO.h"
#include	"KeggGene.h"
#include	"USCG.h"

using namespace std;

map<int, int>* parseGiTaxidFile();
map<int, TaxonomyNode>* parseRootmapFile();
map<string, KO>* createKoMap();
map<string, USCG*>* createUscgMap(map<string, KO>*);

map<string, KeggGene*>* createKeggGeneMap(map<string, KO>*, map<string, USCG*>*, map<int, int>*);

pair<int, map<string, vector<BlastResult*> >*> parseBlastResult();
pair<int, string> splitTaxData(string);

void outputStatus(const string, const string);
void outputStatus(const string, const int, const string);

void normalize(map<string, vector<BlastResult*> >*, map<string, KeggGene*>*, map<string, USCG*>*);
pair<int, BlastResult*> getTopHit(vector<BlastResult*>*);

void countGenes(map<string, vector<BlastResult*> >*, map<string, KeggGene*>*);
double countUscgGenes(map<string, USCG*>*);
void createNormalizeResult(map<string, KeggGene*>*, double);
string showCountSummary(map<int, map<int, map<int, double> > >*);
void showNormalizedCountSummary(map<string, KeggGene*>*);
string createTaxonomyInformation(int, bool);

map<int, TaxonomyNode>* taxonomyRootMap;

ParameterObj param;

int main(const int argc, const char *argv[]) {
	param.ParseParameter(argc, argv);

	// パラメータのパース中にエラーが発生した場合、使い方を表示して終了する.
	if(!param.IsValid()) {
		cout << param.GetUsage(string(argv[0])) << endl;
		return 1;
	}

	outputStatus("START", "Start KEGG Analyzer");

	// gi_taxid_prot.dmpをパースする
	outputStatus("START", "Loading Gi-Taxid map file...");
	map<int, int> *giTaxidMap = parseGiTaxidFile();
	if((*giTaxidMap).empty()) {
		cout << "There are no gene in \"" << param.getFileName(ParameterObj::GI_TAXID) << "\"" << endl;
	}
	outputStatus("END", (*giTaxidMap).size(), " genes loaded.");

	// root_map.txtをパースする
	outputStatus("START", "Loading Taxonomy root map file...");
	taxonomyRootMap = parseRootmapFile();
	if((*taxonomyRootMap).empty()) {
		cout << "There are no species in \"" << param.getFileName(ParameterObj::ROOT_MAP) << "\"" << endl;
	}
	outputStatus("END", (*taxonomyRootMap).size(), " species loaded.");

	// ko_enzyme.listをパースし、KOオブジェクトのmapを生成する.
	outputStatus("START", "Loading KO-Enzyme map file...");
	map<string, KO> *koMap = createKoMap();
	if((*koMap).empty()) {
		cout << "There are no enzyme in \"" << param.getFileName(ParameterObj::KO_ENZYME) << "\"" << endl;
	}
	outputStatus("END", (*koMap).size(), " KOs loaded.");

	// uscg.listをパースし、USCGオブジェクトのmapを生成する.
	outputStatus("START", "Loading USCG map file...");
	map<string, USCG*>* uscgMap = createUscgMap(koMap);
	if((*uscgMap).empty()) {
		cout << "There are no USCG KO in \"" << param.getFileName(ParameterObj::USCG_LIST) << "\"" << endl;
		return 1;
	}
	outputStatus("END", (*uscgMap).size(), " USCGs loaded.");

	// genes_legth.listをパースし、KEGG geneオブジェクトのmapを生成する.
	outputStatus("START", "Loading KEGG genes file...");
	map<string, KeggGene*> *keggGeneMap = createKeggGeneMap(koMap, uscgMap, giTaxidMap);
	if((*keggGeneMap).empty()) {
		cout << "There are no species in \"" << param.getFileName(ParameterObj::GENES_LIST) << "\"" << endl;
		return 1;
	}
	outputStatus("END", (*keggGeneMap).size(), " genes loaded.");

	delete giTaxidMap;

	// blast_result.txtをパースする
	outputStatus("START", "Loading Blast result...");
	pair<int, map<string, vector<BlastResult*> >*> parseResultPair = parseBlastResult();
	map<string, vector<BlastResult*> >* blastResultMap = parseResultPair.second;
	if((*blastResultMap).empty()) {
		cout << "There are no results in \"" << param.getFileName(ParameterObj::BLAST_RESULT) << "\"" << endl;
	}
	outputStatus("END", parseResultPair.first, " blast results loaded.");

	// normalize処理を行う
	outputStatus("START", "Normalizing...");
	normalize(blastResultMap, keggGeneMap, uscgMap);
	outputStatus("END", "Normalize.");

	outputStatus("END", "End KEGG Analyzer");

	return 0;
}

inline void outputStatus(const string status, const string message) {
	time_t current;
	struct tm *local;

	time(&current);
	local = localtime(&current);

	fprintf(stdout, "%04d/%02d/%02d %02d:%02d:%02d\t%s\t%s\n",
			local->tm_year+1900, local->tm_mon+1, local->tm_mday,
			local->tm_hour, local->tm_min, local->tm_sec,
			status.c_str(), message.c_str());
}

inline void outputStatus(const string status, const int num, const string message) {
	stringstream ss;

	ss << num << message;

	outputStatus(status, ss.str());
}

/**
 * gi_taxid_prot.dmp をパースする.
 */
inline map<int, int>* parseGiTaxidFile() {
	ifstream input(param.getFileName(ParameterObj::GI_TAXID).c_str());

	map<int, int> *giTaxidMap = new map<int, int>;

	string line;
	int gi, taxid;
	stringstream ss;

	// 全行ロード
	while(!input.eof() && getline(input, line)) {
		// 行頭に"#"はコメント行としてスキップ
		if(line.find("#") == 0) {
			continue;
		}

		// stringstreamへ読み込み for 行パース
		// gi_taxid_prot.dmpのフォーマットは、　gi[TAB]tax_id
		// 各カラムは数字なのでstringstreamでの分割が可能
		ss.str(line);
		ss.clear(stringstream::goodbit);
		ss >> gi >> taxid;

		// giをkey, tax_idをvalueとするマップに格納
		(*giTaxidMap).insert(pair<int, int>(gi, taxid));
	}

	// パースサマリ
	return giTaxidMap;
}

/**
 * root_map.txt をパースする.
 *  with_name_resolvedに対応済み
 */
inline map<int, TaxonomyNode>* parseRootmapFile() {
	ifstream input(param.getFileName(ParameterObj::ROOT_MAP).c_str());

	map<int, TaxonomyNode> *taxonomyRootMap = new map<int, TaxonomyNode>;

	string line;

	// 行ごとのスプリッタの準備
	Splitter lineSplitter;
	vector<string> delimVector;
	delimVector.push_back(PARSE_DELIMITER);
	delimVector.push_back(PARSE_DELIMITER_LINE_END);

	// スプリット後のデータ
	vector<string> columns;

	// 格納用データ
	pair<int, string> taxDataPair;

	// 全行ロード
	while(!input.eof() && getline(input, line)) {
		// 行頭に"#"はコメント行としてスキップ
		if(line.find("#") == 0) {
			continue;
		}

		// root_map.txtのフォーマットは、　tax_id[TAB]root_tax_id[TAB]phylum_tax_id[TAB]genus_tax_id([TAB]...)
		//  →with_name_resolved版では、各TaxIdの後に :tax_name が付く
		// 各カラム内に空白を含む可能性があるためstringstreamでの分割が不可能
		//  →自作スプリッタSplitterの使用
		lineSplitter.setStr(line, delimVector);
		columns = lineSplitter.split();

		TaxonomyNode node;

		// 1カラム目は自身のノードの情報
		taxDataPair = splitTaxData(columns[0]);
		node.SetId(taxDataPair.first);
		node.SetName(taxDataPair.second);

		// 2カラム目はルートノードの情報
		taxDataPair = splitTaxData(columns[1]);
		node.SetRootId(taxDataPair.first);
		node.SetRootName(taxDataPair.second);

		// 3カラム目は門ノードの情報
		taxDataPair = splitTaxData(columns[2]);
		node.SetPhylumId(taxDataPair.first);
		node.SetPhylumName(taxDataPair.second);

		// 4カラム目は属ノードの情報
		taxDataPair = splitTaxData(columns[3]);
		node.SetGenusId(taxDataPair.first);
		node.SetGenusName(taxDataPair.second);

		// 格納
		(*taxonomyRootMap).insert(pair<int, TaxonomyNode>(node.GetId(), node));
	}

	return taxonomyRootMap;
}

/**
 * tax_id(:tax_node_name)のパースを行い、pairで返す
 */
inline pair<int, string> splitTaxData(string data) {
	// プログラム内で1つで十分
	static Splitter colonSplitter;

	// 区切り文字は":"
	colonSplitter.setStr(data, ":");

	// 最初のカラムはID(整数)
	string id = colonSplitter.next();

	// それ以降は名称だが、名称内に":"を含んでいると嫌なので
	// 「IDとセパレータ(:)とそれ以外」に分割する
	string name = colonSplitter.next();
	if(!name.empty()) {
		name = data.substr(id.length()+1);
	}

	// pairを作成する
	return pair<int, string>(atoi(id.c_str()), name);
}

/**
 * ko_enzyme.list をパースする.
 */
inline map<string, KO>* createKoMap() {
	ifstream input(param.getFileName(ParameterObj::KO_ENZYME).c_str());

	map<string, KO> *koMap = new map<string, KO>;

	string line;
	string ko, enzyme;
	stringstream ss;

	// 全行ロード
	while(!input.eof() && getline(input, line)) {
		// 行頭に"#"はコメント行としてスキップ
		if(line.find("#") == 0) {
			continue;
		}

		// stringstreamへ読み込み for 行パース
		// gi_taxid_prot.dmpのフォーマットは、　gi[TAB]tax_id
		// 各カラムは数字なのでstringstreamでの分割が可能
		ss.str(line);
		ss.clear(stringstream::goodbit);
		ss >> ko >> enzyme;

		if(koMap->find(ko) == koMap->end()) {
			koMap->insert(pair<string, KO>(ko, KO(ko)));
		}

		(*koMap)[ko].AddEnzyme(enzyme);
	}

	return koMap;
}

/**
 * ko_enzyme.list をパースする.
 */
inline map<string, USCG*>* createUscgMap(map<string, KO> *koMap) {
	ifstream input(param.getFileName(ParameterObj::USCG_LIST).c_str());

	map<string, USCG*> *uscgMap = new map<string, USCG*>;

	string line;
	string def, ko;
	stringstream ss;

	// 全行ロード
	while(!input.eof() && getline(input, line)) {
		// 行頭に"#"はコメント行としてスキップ
		if(line.find("#") == 0) {
			continue;
		}

		// stringstreamへ読み込み for 行パース
		// gi_taxid_prot.dmpのフォーマットは、　gi[TAB]tax_id
		// 各カラムは数字なのでstringstreamでの分割が可能
		ss.str(line);
		ss.clear(stringstream::goodbit);
		ss >> def >> ko;

		if(ko.find("ko:") == string::npos) {
			ko = "ko:" + ko;
		}

		(*uscgMap).insert(pair<string, USCG*>(ko, new USCG(ko, def)));
	}

	return uscgMap;
}

/**
 * genes.listをパースし、KeggGeneオブジェクトを生成し、mapにする.
 */
inline map<string, KeggGene*>* createKeggGeneMap(
		map<string, KO>* koMap, map<string, USCG*>* uscgMap, map<int, int>* giTaxidMap) {

	ifstream input(param.getFileName(ParameterObj::GENES_LIST).c_str());
	map<string, KeggGene*> *keggGeneMap = new map<string, KeggGene*>;

	string line;
	string id, ko;
	int len, gi;
	stringstream ss;

	Splitter lineSplitter, koSplitter;

	// 全行ロード
	while(!input.eof() && getline(input, line)) {
		// 行頭に"#"はコメント行としてスキップ
		if(line.find("#") == 0) {
			continue;
		}

		// genes_ncbi-gi.listのフォーマットは、　kegg_id[TAB]gi
		// 各カラムはタブ及びスペースを含まないが、空のカラムもあるのでstringstreamでの分割が不可能
		lineSplitter.setStr(line, "\t");
		vector<string> columns = lineSplitter.split();
		id = columns[0];
		len = atoi(columns[1].c_str());
		gi = atoi(columns[2].c_str());
		ko = columns[3];

		KeggGene *gene = new KeggGene(id, len);

		// set Taxonomy
		(*gene).SetTaxonomy(gi, giTaxidMap, taxonomyRootMap);

		// set KO
		(*gene).SetKeggOrtholog(ko, koMap, uscgMap);

		(*keggGeneMap).insert(pair<string, KeggGene*>(id, gene));
	}

	return keggGeneMap;
}

/**
 * blast_result.txt をパースする.
 */
inline pair<int, map<string, vector<BlastResult*> >*> parseBlastResult() {
	ifstream input(param.getFileName(ParameterObj::BLAST_RESULT).c_str());

	// blastの結果は、1つのIDに対して複数存在するため、multimapで実装する
	// ※map<string, vector<BlastResult> > もありだが、この場合はこれで十分
	map<string, vector<BlastResult*> > *blastResultMap = new map<string, vector<BlastResult*> >;

	stringstream ss;

	string line;

	// blast_result.txtから取得できるデータの格納変数
	string queryId, ids;
	double identity, evalue, score;
	int alignmentLength, mismatch, gap;
	int qStart, qEnd, sStart, sEnd;

	Splitter splitter;

	int count = 0;

	// 全行ロード
	while(!input.eof() && getline(input, line)) {
		// stringstreamへ読み込み for 行パース
		// blast_result.txtのフォーマットは、各カラムが[TAB]で区切られた形
		// 各カラムは数字なのでstringstreamでの分割が可能

		// ghostxの出力は2カラム目に半角スペースを含むためsstreamの分割が不可能
		splitter.setStr(string(line), "\t");
		vector<string> columns = splitter.split();

		ss.str("");
		ss.clear(stringstream::goodbit);
		ss << columns[0] << " " << columns[2] << " " << columns[3] << " " <<
			columns[4] << " " << columns[5] << " " << columns[6] << " " <<
			columns[7] << " " << columns[8] << " " << columns[9] << " " <<
			columns[10] << " " << columns[11];
		ss >> queryId >> identity >> alignmentLength >> mismatch >> gap >>
				qStart >> qEnd >> sStart >> sEnd >> evalue >> score;

		BlastResult *result = new BlastResult();

		// 2カラム目はgiなど
		(*result).setId(columns[1]);

		// 3カラム目は% identity
		(*result).setIdentity(identity);

		// 4カラム目はalignment length
		(*result).setAlignmentLength(alignmentLength);

		// 5カラム目はmismatch
		(*result).setMismatches(mismatch);

		// 6カラム目はgap openings
		(*result).setGap(gap);

		// 7カラム目はquery start
		// 8カラム目はquery end
		// 9カラム目はsubject start
		// 10カラム目はsubject end

		// 11カラム目はe-value
		(*result).setEvalue(evalue);

		// 12カラム目はbit score
		(*result).setScore(score);

		(*blastResultMap)[queryId].push_back(result);

		count++;
	}

	return pair<int, map<string, vector<BlastResult*> >*>(count, blastResultMap);
}

/**
 * Kegg Analyzerの本体
 */
inline void normalize(map<string, vector<BlastResult*> >* blastResultMap, map<string, KeggGene*>* keggGeneMap, map<string, USCG*>* uscgMap) {

	// 出力ファイル群の準備

	ofstream *file;

// 1. count-up

	outputStatus("START", "1. count genes...");
	// 各レベルでのカウンタ
	countGenes(blastResultMap, keggGeneMap);
	outputStatus("END",  "done.");

// 2. USCG count

	outputStatus("START", "2. count USCGs...");
	// USCGsの平均値
	double mean = countUscgGenes(uscgMap);
	outputStatus("END",  "done.");

// 3. normalize

	outputStatus("START", "3. normalizing...");
	createNormalizeResult(keggGeneMap, mean);
	outputStatus("END", "done.");
}

inline void countGenes(map<string, vector<BlastResult*> >* blastResultMap, map<string, KeggGene*>* keggGeneMap) {
	map<int, map<int, map<int, double> > > *countByEachLevel = new map<int, map<int, map<int, double> > >;

	int insufficientResult = 0;
	int sufficientResult = 0;
	int taxKeggIdMiss = 0;
	int totalResultCount = 0;
	int keggGeneCount = 0;
	int searchResultCount = 0;

	ofstream normalizeResultFile(param.getFileName(ParameterObj::BEFORE_NORMALIZE).c_str());
	ofstream normalizeErrorFile(param.getFileName(ParameterObj::NORMALIZE_ERROR).c_str());

	stringstream notFoundNcbiGene;
	stringstream notFoundKeggGene;
	stringstream insufficientQuality;

	for(map<string, vector<BlastResult*> >::iterator itr = (*blastResultMap).begin(); itr != (*blastResultMap).end(); itr++) {
		string queryId = itr->first;
		totalResultCount += (itr->second).size();
		pair<int, BlastResult*> topHitResult = getTopHit(&(itr->second));
		BlastResult* topHit = topHitResult.second;
		int status = topHitResult.first;

		/* status == 1 means read has blast result. */
		if(status == 1) {
			/* topHit != NULL means top hit result of read satisfies threshold. */
			if(topHit != NULL) {
				if(keggGeneMap->find(topHit->getId()) != keggGeneMap->end()) {
					KeggGene* gene = (*keggGeneMap)[topHit->getId()];

					TaxonomyNode *node = gene->GetTaxNode();
					if(node != NULL) {
						gene->inc();

						(*countByEachLevel)[node->GetRootId()][node->GetPhylumId()][node->GetGenusId()]++;
					} else {
						notFoundNcbiGene << queryId << endl;
						taxKeggIdMiss++;
					}

					keggGeneCount++;
				} else {
					notFoundKeggGene << queryId << endl;
				}

				sufficientResult++;
			} else {
				insufficientQuality << queryId << endl;

				insufficientResult++;
			}

			searchResultCount++;
		}
	}

	normalizeResultFile << "# count result of aligned reads on NCBI Taxonomy tree." << endl;
	normalizeResultFile << endl;
	normalizeResultFile << "Total Homology search vs KEGG result:\t" << totalResultCount << endl;
	normalizeResultFile << endl;
	normalizeResultFile << "Total Top Hit Result:\t" << searchResultCount << endl;
	normalizeResultFile << "\tScore>=" << param.getBlastThresholdScore() << " and Identity>=" << param.getBlastThresholdIdentity() << " Result:\t" << sufficientResult << endl;
	normalizeResultFile << "\tNot Enough Score or Identity Result:\t" << insufficientResult << endl;
	normalizeResultFile << endl;
	normalizeResultFile << "Total reads aligned to KEGG genes:\t" << keggGeneCount << endl;

	normalizeResultFile << showCountSummary(countByEachLevel) << endl;;

	normalizeErrorFile << "# These reads contain errors." << endl;
	normalizeErrorFile << endl;
	normalizeErrorFile << "# 1. Homology search hits, satisfies quality, but KEGG gene not assigned to NCBI gene." << endl;
	normalizeErrorFile << notFoundNcbiGene.str() << endl;
	normalizeErrorFile << "# 2. Homology search hits, satisfies quality, but can't find KEGG gene detail." << endl;
	normalizeErrorFile << notFoundKeggGene.str() << endl;
	normalizeErrorFile << "# 3. Homology search hits, but not satisfies quality." << endl;
	normalizeErrorFile << insufficientQuality.str() << endl;

	delete countByEachLevel;
}

/**
 *
 */
inline pair<int, BlastResult*> getTopHit(vector<BlastResult*>* blastResultVec) {
	ofstream normalizeErrorFile(param.getFileName(ParameterObj::NORMALIZE_ERROR).c_str());

	if((*blastResultVec).size() > 0) {
		BlastResult *topHit = (*blastResultVec)[0];

		if((*topHit).getIdentity() < param.getBlastThresholdIdentity() || (*topHit).getScore() < param.getBlastThresholdScore()) {
			return pair<int, BlastResult*>(1, NULL);
		} else {
			return pair<int, BlastResult*>(1, topHit);
		}
	} else {
		return pair<int, BlastResult*>(0, NULL);
	}
}

/**
 * 各レベルでのカウントのサマリを表示する.
 * ツリー状に情報を保持しているため、その順に書き出す.
 * 関数が大きいので一部分割したい.
 */
inline string showCountSummary(map<int, map<int, map<int, double> > >* rootCountMap) {
	map<int, map<int, map<int, double> > >::iterator rootIterator = (*rootCountMap).begin();

	stringstream rootStream;

	rootStream << "----- summary -----" << endl;
	rootStream << "# tax_id, name, superkingdom, phylum, genus, count" << endl;

	// rootレベルでの出力
	while(rootIterator != (*rootCountMap).end()) {
		map<int, map<int, double> > *phylumCountMap = &((*rootIterator).second);
		map<int, map<int, double> >::iterator phylumIterator = (*phylumCountMap).begin();

		stringstream phylumStream;

		double rootSum = 0;
		// phylumレベルでの出力
		while(phylumIterator != (*phylumCountMap).end()) {
			map<int, double> *genusCountMap = &((*phylumIterator).second);
			map<int, double>::iterator genusIterator = (*genusCountMap).begin();

			stringstream genusStream;

			double phylumSum = 0;
			// genusレベルでの出力
			while(genusIterator != (*genusCountMap).end()) {
				// genusStream << "   " << createTaxonomyInformation((*genusIterator).first, true) << "\t" << (*genusIterator).second << LINE_DELIMITER;
				genusStream << (*genusIterator).first << "\t" << createTaxonomyInformation((*rootIterator).first, false) << "\t" << createTaxonomyInformation((*phylumIterator).first, false) << "\t" << createTaxonomyInformation((*genusIterator).first, false) << "\t" << (*genusIterator).second << endl;

				// phylumのカウントはgenusのカウントの総和
				phylumSum += (*genusIterator).second;
				genusIterator++;
			}

			// phylumの情報を出力してから
			// phylumStream << "  " << createTaxonomyInformation((*phylumIterator).first, true) << "\t" << phylumSum << LINE_DELIMITER;
			phylumStream << (*phylumIterator).first << "\t" << createTaxonomyInformation((*rootIterator).first, false) << "\t" << createTaxonomyInformation((*phylumIterator).first, false) << "\t" << "-" << "\t" << phylumSum << endl;

			// genusの情報を出力する
			// カウントアップの都合上この処理になる
			phylumStream << genusStream.str();

			// rootのカウントはphylumのカウントの総和
			rootSum += phylumSum;

			phylumIterator++;
		}

		// rootの情報を出力してから
		// rootStream << " " << createTaxonomyInformation((*rootIterator).first, true) << "\t" << rootSum << LINE_DELIMITER;
		rootStream << createTaxonomyInformation((*rootIterator).first, true) << "\t" << "-" << "\t" << "-" << "\t" << rootSum << endl;

		// phylumの情報(=genusの情報を含む)を出力する
		// カウントアップの都合上この処理になる
		rootStream << phylumStream.str();

		rootIterator++;
	}

	return rootStream.str();
}

inline double countUscgGenes(map<string, USCG*>* uscgMap) {
	ofstream uscgCountFile(param.getFileName(ParameterObj::USCG_COUNT).c_str());

	uscgCountFile << "# USCG(Universal Single Copy Genes) count result" << endl;
	uscgCountFile << endl;

	map<string, int> uscgCountMap;
	vector<int> countVec;
	int totalCount = 0;

	stringstream str;

	for(map<string, USCG*>::iterator itr = uscgMap->begin(); itr != uscgMap->end(); itr++) {
		string koId = itr->first;
		USCG *uscg = itr->second;

		vector<KeggGene*> *geneVec = uscg->GetGeneVec();

		stringstream eachStr;

		int count = 0;
		for(vector<KeggGene*>::iterator itr2 = geneVec->begin(); itr2 != geneVec->end(); itr2++) {
			KeggGene *gene = *itr2;

			int _count = gene->GetCount();
			if(_count > 0) {
				eachStr << "\t" << gene->GetId() << "\t" << _count << endl;

				count += _count;
			}
		}

		if(count > 0) {
			countVec.push_back(count);
			totalCount += count;

			str << uscg->GetDef() << "(" << koId << ")\t" << count << endl;
			str << eachStr.str();
		}
	}

	double mean = 0;
	if(countVec.size() > 0) {
		sort(countVec.begin(), countVec.end());

		mean = totalCount / countVec.size();
		double median =
			(countVec.size() % 2 == 0 ?
				(countVec[countVec.size() / 2 - 1] + countVec[countVec.size() / 2 + 1]) / 2 :
				countVec[(int)(countVec.size() / 2)]);
		double std = 0;
		{
			double diff = 0;
			for(vector<int>::iterator itr = countVec.begin(); itr != countVec.end(); itr++) {
				diff += pow((*itr)-mean, 2);
			}
			std = sqrt(diff/(countVec.size()-1));
		}

		uscgCountFile <<
			"# USCG count=" << countVec.size() << ", min=" << countVec[0] << ", max=" << countVec[countVec.size()-1] <<
			", mean=" << mean << ", median=" << median << ", std=" << std << endl;
		uscgCountFile << str.str();
	} else {
		uscgCountFile << "# There are no USCGs." << endl;
	}

	return mean;
}

inline void createNormalizeResult(map<string, KeggGene*>* keggGeneMap, double mean) {
	ofstream normalizeCountFile(param.getFileName(ParameterObj::NORMALIZE_COUNT).c_str());

	normalizeCountFile << "# Normalize count result file for KeGG" << endl;
	normalizeCountFile << endl;
	normalizeCountFile << "# kegg_id, count, count/gene_length, count/gene_length/mean(uscg), ec, ko, genus" << endl;

	if(mean == 0) {
		normalizeCountFile << "# There are no USCGs in target reads." << endl;
		return;
	}

	for(map<string, KeggGene*>::iterator itr = keggGeneMap->begin(); itr != keggGeneMap->end(); itr++) {
		string keggId = itr->first;
		KeggGene* gene = itr->second;

		int count = gene->GetCount();
		if(count > 0 && gene->GetLength() > 0) {
			string koStr, ecStr;

			vector<string> alreadyPutEcVec;

			if(gene->GetKoVec() != NULL) {
				for(vector<KO *>::iterator itr2 = gene->GetKoVec()->begin(); itr2 != gene->GetKoVec()->end(); itr2++) {
					KO* ko = *itr2;

					if(koStr.size() > 0) {
						koStr += + ',';
					}
					koStr += ko->GetId();

					if(ko->GetEnzymeVec() != NULL) {
						for(vector<string>::iterator itr3 = ko->GetEnzymeVec()->begin(); itr3 != ko->GetEnzymeVec()->end(); itr3++) {
							if(find(alreadyPutEcVec.begin(), alreadyPutEcVec.end(), *itr3) == alreadyPutEcVec.end()) {
								string ec = *itr3;

								alreadyPutEcVec.push_back(ec);

								if(ecStr.size() > 0) {
									if(ecStr == "EC not found") {
										ecStr = "";
									} else {
										ecStr += ",";
									}
								}
								ecStr += ec;
							}
						}
					} else {
						if(ecStr.size() == 0) {
							ecStr = "EC not found";
						}
					}
				}
			} else {
				koStr = "KO not found";
				ecStr = "EC not found";
			}

			gene->SetNorm((double)count / gene->GetLength() / mean);

			normalizeCountFile << keggId << "\t" <<
				count << "\t" << (double)count/gene->GetLength() << "\t" << (double)count/gene->GetLength()/mean << "\t" <<
				ecStr << "\t" << koStr << "\t" << gene->GetTaxNode()->GetGenusName() << endl;
		}
	}

	showNormalizedCountSummary(keggGeneMap);
}

/**
 * 各レベルでのカウントのサマリを作成する.
 */
inline void showNormalizedCountSummary(map<string, KeggGene*>* keggGeneMap) {
	ofstream normalizeCountFile(param.getFileName(ParameterObj::NORMALIZE_OUT).c_str());

	normalizeCountFile << "# normalize result arrangement to NCBI Taxonomy" << endl;
	normalizeCountFile << endl;

	map<string, KeggGene*>::iterator itr = (*keggGeneMap).begin();

	map<int, map<int, map<int, double> > > *countByEachLevel = new map<int, map<int, map<int, double> > >;

	map<string, int> countSpeciesMap;
	map<string, int> countGenesMap;

	while(itr != (*keggGeneMap).end()) {
		string keggId = itr->first;
		KeggGene* gene = itr->second;
		TaxonomyNode* node = gene->GetTaxNode();

		if(gene->GetNorm() > 0 && node != NULL) {
			(*countByEachLevel)[node->GetRootId()][node->GetPhylumId()][node->GetGenusId()] += gene->GetNorm();

			string species = keggId.substr(0, keggId.find(":"));
			countSpeciesMap[species]++;
			countGenesMap[keggId]++;
		}

		itr++;
	}

	normalizeCountFile << "Total KEGG species:\t" << countSpeciesMap.size() << endl;
	normalizeCountFile << "Total KEGG genes:\t" << countGenesMap.size() << endl;

	normalizeCountFile << showCountSummary(countByEachLevel) << endl;

	delete countByEachLevel;
}

/**
 * レジェンドの出力
 */
inline string createTaxonomyInformation(int taxid, bool isShowTaxId) {
	map<int, TaxonomyNode>::iterator taxIterator = taxonomyRootMap->find(taxid);

	string nodeName;

	// 指定するtax_idからTaxonomy Nodeの情報を取得
	if(taxIterator != taxonomyRootMap->end()) {
		// もしあれば名前取得
		TaxonomyNode node = (*taxonomyRootMap)[taxid];
		nodeName = node.GetName();
	} else {
		// なければundefined
		nodeName = "undefined";
	}

	// プログラム内で1つで十分
	static stringstream ss;
	ss.clear(stringstream::goodbit);
	ss.str("");

	if(isShowTaxId) {
		// Taxonomy IDと名前は[TAB]区切りで出力
		// 名前が定義されていなくても空欄を出力
		ss << taxid << "\t" << nodeName;
	} else {
		ss << nodeName;
	}

	return ss.str();
}
